# -*- coding: utf-8 -*-
#
#  Copyright (C) 2001, 2002 by Tamito KAJIYAMA
#  Copyright (C) 2002, 2003 by MATSUMURA Namihiko <nie@counterghost.net>
#  Copyright (C) 2002-2012 by Shyouzou Sugitani <shy@users.sourceforge.jp>
#  Copyright (C) 2003 by Shun-ichi TAHARA <jado@flowernet.gr.jp>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License (version 2) as
#  published by the Free Software Foundation.  It is distributed in the
#  hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the GNU General Public License for more details.
#

import gtk
import pango
import cairo
import glib

import ninix.pix


class Balloon(object):

    def __init__(self):
        self.request_parent = lambda *a: None # dummy
        self.synchronized = []
        self.user_interaction = False
        self.window = []
        # create communicatebox
        self.communicatebox = CommunicateBox()
        self.communicatebox.set_responsible(self.handle_request)
        # create teachbox
        self.teachbox = TeachBox()
        self.teachbox.set_responsible(self.handle_request)
        # create inputbox
        self.inputbox = InputBox()
        self.inputbox.set_responsible(self.handle_request)

    def set_responsible(self, request_method):
        self.request_parent = request_method

    def handle_request(self, event_type, event, *arglist, **argdict):
        assert event_type in ['GET', 'NOTIFY']
        handlers = {
            'reset_user_interaction': self.reset_user_interaction,
            }
        handler = handlers.get(event)
        if handler is None:
            result = self.request_parent(
                event_type, event, *arglist, **argdict)
        else:
            result = handler(*arglist, **argdict)
        if event_type == 'GET':
            return result

    def reset_user_interaction(self):
        self.user_interaction = False

    def get_text_count(self, side):
        if len(self.window) > side:
            return self.window[side].get_text_count()
        else:
            return 0

    def get_window(self, side):
        if len(self.window) > side:
            return self.window[side].window ## FIXME
        else:
            return None

    def reset_text_count(self, side):
        if len(self.window) > side:
            self.window[side].reset_text_count()

    def reset_balloon(self):
        for balloon_window in self.window:
            balloon_window.reset_balloon()

    def create_gtk_window(self, title):
        window = ninix.pix.TransparentWindow()
        window.set_focus_on_map(False)
        window.set_title(title)
        window.set_decorated(False)
        window.set_resizable(False)
        window.set_skip_pager_hint(False)
        window.set_skip_taskbar_hint(True)
        window.connect('delete_event', self.delete)
        window.realize()
        return window

    def identify_window(self, win):
        for balloon_window in self.window:
            if win == balloon_window.window.window:
                return True
        return False

    def delete(self, window, event):
        return True

    def finalize(self):
        for balloon_window in self.window:
            balloon_window.destroy()
        self.window = []
        self.communicatebox.destroy()
        self.teachbox.destroy()
        self.inputbox.destroy()

    def new(self, desc, balloon):
        self.desc = desc
        self.directory = balloon['balloon_dir'][0]
        balloon0 = {}
        balloon1 = {}
        communicate0 = None
        communicate1 = None
        communicate2 = None
        communicate3 = None
        for key, value in balloon.items():
            if key in ['arrow0', 'arrow1']:
                balloon0[key] = value
                balloon1[key] = value
            elif key == 'sstp':
                balloon0[key] = value  # sstp marker
            elif key.startswith('s'):
                balloon0[key] = value  # Sakura
            elif key.startswith('k'):
                balloon1[key] = value  # Unyuu
            elif key == 'c0':
                communicate0 = value # send box
            elif key == 'c1':
                communicate1 = value # communicate box
            elif key == 'c2':
                communicate2 = value # teach box
            elif key == 'c3':
                communicate3 = value # input box
        self.balloon0 = balloon0
        self.balloon1 = balloon1
        # create balloon windows
        for balloon_window in self.window:
            balloon_window.destroy()
        self.window = []
        self.add_window(0)
        self.add_window(1)
        # configure communicatebox
        self.communicatebox.new(desc, communicate1)
        # configure teachbox
        self.teachbox.new(desc, communicate2)
        # configure inputbox
        self.inputbox.new(desc, communicate3)

    def add_window(self, side):
        assert len(self.window) == side
        if side == 0:
            name = 'balloon.sakura'
            id_format = 's{0:d}'
            balloon = self.balloon0
        elif side == 1:
            name = 'balloon.kero'
            id_format = 'k{0:d}'
            balloon = self.balloon1
        else:
            name = 'balloon.char{0:d}'.format(side)
            id_format = 'k{0:d}'
            balloon = self.balloon1
        gtk_window = self.create_gtk_window(name)
        balloon_window = BalloonWindow(
            gtk_window, side, self.desc, balloon,
            id_format)
        balloon_window.set_responsible(self.handle_request)
        self.window.append(balloon_window)

    def reset_fonts(self):
        for window in self.window:
            window.reset_fonts()

    def get_balloon_directory(self):
        return self.directory

    def get_balloon_size(self, side):
        if len(self.window) > side:
            return self.window[side].get_balloon_size()
        else:
            return (0, 0)

    def get_balloon_windowposition(self, side):
        if len(self.window) > side:
            return self.window[side].get_balloon_windowposition()
        else:
            return (0, 0)

    def set_balloon_default(self):
        for side in range(len(self.window)):
            self.window[side].set_balloon(0) ## FIXME

    def set_balloon(self, side, num):
        if len(self.window) > side:
            self.window[side].set_balloon(num)

    def set_direction(self, side, dir):
        if len(self.window) > side:
            self.window[side].set_direction(dir)

    def set_position(self, side, base_x, base_y):
        if len(self.window) > side:
            self.window[side].set_position(base_x, base_y)

    def get_position(self, side):
        if len(self.window) > side:
            return self.window[side].get_position()
        else:
            return (0, 0)

    def is_shown(self, side):
        return 0 if len(self.window) <= side else self.window[side].is_shown()

    def show(self, side):
        if len(self.window) > side:
            self.window[side].show()

    def hide_all(self):
        for side in range(len(self.window)):
            self.window[side].hide()

    def hide(self, side):
        if len(self.window) > side:
            self.window[side].hide()

    def raise_all(self):
        for side in range(len(self.window)):
            self.window[side].raise_()

    def raise_(self, side):
        if len(self.window) > side:
            self.window[side].raise_()

    def lower_all(self):
        for side in range(len(self.window)):
            self.window[side].lower()

    def lower(self, side):
        if len(self.window) > side:
            self.window[side].lower()

    def synchronize(self, list):
        self.synchronized = list

    def clear_text_all(self):
        for side in range(len(self.window)):
            self.clear_text(side)

    def clear_text(self, side):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].clear_text()
        else:
            if len(self.window) > side:
                self.window[side].clear_text()

    def append_text(self, side, text):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].append_text(text)
        else:
            if len(self.window) > side:
                self.window[side].append_text(text)

    def append_sstp_marker(self, side):
        if len(self.window) > side:
            self.window[side].append_sstp_marker()

    def append_link_in(self, side, label):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].append_link_in(label)
        else:
            if len(self.window) > side:
                self.window[side].append_link_in(label)

    def append_link_out(self, side, label, value):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].append_link_out(label, value)
        else:
            if len(self.window) > side:
                self.window[side].append_link_out(label, value)

    def append_link(self, side, label, value, newline_required=0):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].append_link_in(label)
                    self.window[side].append_text(value)
                    self.window[side].append_link_out(label, value)
                    if newline_required:
                        self.window[side].set_newline()
        else:
            if len(self.window) > side:
                self.window[side].append_link_in(label)
                self.window[side].append_text(value)
                self.window[side].append_link_out(label, value)
                if newline_required:
                    self.window[side].set_newline()

    def append_meta(self, side, tag):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].append_meta(tag)
        else:
            if len(self.window) > side:
                self.window[side].append_meta(tag)

    def append_image(self, side, path, x, y):
        if len(self.window) > side:
            self.window[side].append_image(path, x, y)

    def show_sstp_message(self, message, sender):
        self.window[0].show_sstp_message(message, sender)

    def hide_sstp_message(self):
        self.window[0].hide_sstp_message()

    def open_communicatebox(self):
        if not self.user_interaction:
            self.user_interaction = True
            self.communicatebox.show()

    def open_teachbox(self):
        if not self.user_interaction:
            self.user_interaction = True
            self.request_parent('NOTIFY', 'notify_event', 'OnTeachStart')
            self.teachbox.show()

    def open_inputbox(self, symbol, limittime=-1, default=None):
        if not self.user_interaction:
            self.user_interaction = True
            self.inputbox.set_symbol(symbol)
            self.inputbox.set_limittime(limittime)
            self.inputbox.show(default)


class BalloonWindow(object):

    def __init__(self, window, side, desc, balloon, id_format):
        self.window = window
        self.side = side
        self.request_parent = lambda *a: None # dummy
        self.desc = desc
        self.balloon = balloon
        self.balloon_id = None
        self.id_format = id_format
        self.num = 0
        self.__shown = 0
        self.sstp_marker = []
        self.sstp_region = None
        self.sstp_message = None
        self.images = []
        self.width = 0
        self.height = 0
        self.__font_name = ''
        self.text_count = 0
        self.balloon_pixbuf = None
        # create drawing area
        self.darea = gtk.DrawingArea()
        self.darea.set_events(gtk.gdk.EXPOSURE_MASK|
                              gtk.gdk.BUTTON_PRESS_MASK|
                              gtk.gdk.POINTER_MOTION_MASK|
                              gtk.gdk.POINTER_MOTION_HINT_MASK|
                              gtk.gdk.SCROLL_MASK)
        self.darea.connect('expose_event', self.redraw)
        self.darea.connect('button_press_event', self.button_press)
        self.darea.connect('motion_notify_event', self.motion_notify)
        self.darea.connect('scroll_event', self.scroll)
        self.window.add(self.darea)
        self.window.realize()
        self.window.window.set_back_pixmap(None, False)
        self.layout = pango.Layout(self.darea.get_pango_context())
        self.sstp_layout = pango.Layout(self.darea.get_pango_context())
        mask_r = desc.get_with_type('maskcolor.r', int, 128)
        mask_g = desc.get_with_type('maskcolor.g', int, 128)
        mask_b = desc.get_with_type('maskcolor.b', int, 128)
        self.cursor_color = (mask_r / 255., mask_g / 255., mask_b / 255.)
        text_r = desc.get_with_type(['font.color.r', 'fontcolor.r'], int, 0)
        text_g = desc.get_with_type(['font.color.g', 'fontcolor.g'], int, 0)
        text_b = desc.get_with_type(['font.color.b', 'fontcolor.b'], int, 0)
        self.text_normal_color = (text_r / 255., text_g / 255., text_b / 255.)
        if desc.get_with_type('maskmethod', int) == 1:
            text_r = 255 - text_r
            text_g = 255 - text_g
            text_b = 255 - text_b
        self.text_active_color = (text_r / 255., text_g / 255., text_b / 255.)
        # initialize
        self.direction = min(side, 1) ## kluge: multi character
        self.position = (0, 0)
        self.reset_fonts()
        self.clear_text()

    def set_responsible(self, request_method):
        self.request_parent = request_method

    @property
    def scale(self):
        scalling = self.request_parent('GET', 'get_preference', 'balloon_scalling')
        scale = self.request_parent('GET', 'get_preference', 'surface_scale')
        return scale if scalling else 100 # [%]

    def get_balloon_windowposition(self):
        winpos_x = self.__get_with_scaling('windowposition.x', int, 0)
        winpos_y = self.__get_with_scaling('windowposition.y', int, 0)
        return winpos_x, winpos_y

    def get_pixbuf(self, balloon_id):
        #assert balloon_id in self.balloon
        try:
            path, config = self.balloon[balloon_id]
            use_pna = self.request_parent('GET', 'get_preference', 'use_pna')
            pixbuf = ninix.pix.create_pixbuf_from_file(
                path, use_pna=use_pna)
        except:
            return None
        scale = self.scale
        w = pixbuf.get_width()
        h = pixbuf.get_height()
        w = max(8, int(w * scale / 100))
        h = max(8, int(h * scale / 100))
        pixbuf = pixbuf.scale_simple(w, h, gtk.gdk.INTERP_BILINEAR)
        return pixbuf, (w, h)

    def reset_fonts(self):
        font_name = self.request_parent('GET', 'get_preference', 'balloon_fonts')
        if self.__font_name == font_name:
            return
        self.font_desc = pango.FontDescription(font_name)
        pango_size = self.font_desc.get_size()
        if pango_size == 0:
            default_size = 12 # for Windows environment
            size = self.desc.get_with_type(
                ['font.height', 'font.size'], int, default_size)
            pango_size = size * 3 / 4 # convert from Windows to GTK+
            pango_size *= pango.SCALE
        scale = self.scale
        self.font_desc.set_size(pango_size * scale / 100)
        self.__font_name = font_name
        self.layout.set_font_description(self.font_desc)
        self.layout.set_wrap(pango.WRAP_CHAR) # XXX
        # font for sstp message
        if self.side == 0:
            self.sstp_font_desc = pango.FontDescription(font_name)
            pango_size = self.sstp_font_desc.get_size()
            if pango_size == 0:
                default_size = 10 # for Windows environment
                size = self.desc.get_with_type(
                    'sstpmessage.font.height', int, default_size)
                pango_size = size * 3 / 4 # convert from Windows to GTK+
                pango_size *= pango.SCALE
            self.sstp_font_desc.set_size(pango_size * scale / 100)
            self.sstp_layout.set_font_description(self.sstp_font_desc)
            self.sstp_layout.set_wrap(pango.WRAP_CHAR)
        if self.balloon_id is not None:
            self.reset_message_regions()
            if self.__shown:
                self.darea.queue_draw()

    @property
    def alpha_channel(self):
        alpha = self.request_parent('GET', 'get_preference', 'balloon_alpha')
        if alpha is None or not 0.1 <= alpha <= 1.0:
            alpha = 1.0
        return alpha

    def reset_sstp_marker(self):
        if self.side == 0:
            assert self.balloon_pixbuf is not None
            w, h = self.balloon_pixbuf[1]
            # sstp marker position
            self.sstp = []
            x = self.config_adjust('sstpmarker.x', w,  30)
            y = self.config_adjust('sstpmarker.y', h, -20)
            self.sstp.append((x, y)) # sstp marker
            x = self.config_adjust('sstpmessage.x', w,  50)
            y = self.config_adjust('sstpmessage.y', h, -20)
            self.sstp.append((x, y)) # sstp message
        # sstp marker pixbuf (not only for self.side == 0)
        self.sstp_pixbuf = self.get_pixbuf('sstp')

    def reset_arrow(self):
        # arrow positions
        self.arrow = []
        assert self.balloon_pixbuf is not None
        w, h = self.balloon_pixbuf[1]
        x = self.config_adjust('arrow0.x', w, -10)
        y = self.config_adjust('arrow0.y', h,  10)
        self.arrow.append((x, y))
        x = self.config_adjust('arrow1.x', w, -10)
        y = self.config_adjust('arrow1.y', h, -20)
        self.arrow.append((x, y))
        # arrow pixbufs and sizes
        self.arrow0_pixbuf = self.get_pixbuf('arrow0')
        self.arrow1_pixbuf = self.get_pixbuf('arrow1')

    def reset_message_regions(self):
        w, h = self.layout.get_pixel_size()
        self.font_height = h
        self.line_space = 1
        self.layout.set_spacing(self.line_space)
        # font metrics
        origin_x = self.__get_with_scaling(
            'origin.x', int,
            self.__get_with_scaling(
                'zeropoint.x', int,
                self.__get_with_scaling('validrect.left', int, 14)))
        origin_y = self.__get_with_scaling(
            'origin.y', int,
            self.__get_with_scaling(
                'zeropoint.y', int,
                self.__get_with_scaling('validrect.top', int, 14)))
        wpx = self.__get_with_scaling(
            'wordwrappoint.x', int, 
            self.__get_with_scaling('validrect.right', int, -14))
        if wpx > 0:
            line_width = wpx - origin_x
        elif wpx < 0:
            line_width = self.width - origin_x + wpx
        else:
            line_width = self.width - origin_x * 2
        wpy = self.__get_with_scaling('validrect.bottom', int, -14)
        if wpy > 0:
            text_height = min(wpy, self.height) - origin_y
        elif wpy < 0:
            text_height = self.height - origin_y + wpy
        else:
            text_height = self.height - origin_y * 2
        line_height = self.font_height + self.line_space
        self.lines = text_height / line_height
        self.line_regions = []
        y = origin_y
        for _ in range(self.lines + 1):
            self.line_regions.append((origin_x, y, line_width, line_height))
            y = y + line_height
        self.line_width = line_width
        # sstp message region
        if self.side == 0:
            w, h = self.sstp_layout.get_pixel_size()
            x, y = self.sstp[1]
            w = line_width + origin_x - x
            self.sstp_region = (x, y, w, h)

    def update_line_regions(self, offset, new_y):
        origin_y = self.__get_with_scaling(
            'origin.y', int, 
            self.__get_with_scaling(
                'zeropoint.y', int,
                self.__get_with_scaling('validrect.top', int, 14)))
        wpy = self.__get_with_scaling('validrect.bottom', int, -14)
        if wpy > 0:
            text_height = min(wpy, self.height) - origin_y
        elif wpy < 0:
            text_height = self.height - origin_y + wpy
        else:
            text_height = self.height - origin_y * 2
        line_height = self.font_height + self.line_space
        origin_x, y, line_width, line_height = self.line_regions[offset]
        self.lines = offset + (text_height - new_y) / line_height
        y = new_y
        for i in range(offset, len(self.line_regions)):
            self.line_regions[i] = (origin_x, y, line_width, line_height)
            y += line_height
        for i in range(len(self.line_regions), self.lines + 1):
            self.line_regions.append((origin_x, y, line_width, line_height))
            y += line_height

    def get_balloon_size(self):
        return (self.width, self.height)

    def reset_balloon(self):
        self.set_balloon(self.num)

    def set_balloon(self, num):
        self.num = num
        balloon_id = self.id_format.format(num * 2 + self.direction)
        self.balloon_pixbuf = self.get_pixbuf(balloon_id)
        if self.balloon_pixbuf is None:
            balloon_id = self.id_format.format(0 + self.direction)
            self.balloon_pixbuf = self.get_pixbuf(balloon_id)
        assert self.balloon_pixbuf is not None
        self.balloon_id = balloon_id
        # change pixbuf and window position
        x, y = self.position
        self.width, self.height = self.balloon_pixbuf[1]
        self.darea.set_size_request(self.width, self.height)
        mask = gtk.gdk.Pixmap(None, self.width, self.height, 1)
        self.balloon_pixbuf[0].render_threshold_alpha(
            mask, 0, 0, 0, 0, self.width, self.height, 1)
        self.window.mask = mask
        self.reset_arrow()
        self.reset_sstp_marker()
        self.reset_message_regions()
        if self.__shown:
            self.request_parent('NOTIFY', 'position_balloons')
            self.darea.queue_draw()

    def set_direction(self, dir):
        if self.direction != dir:
            self.direction = dir # 0: left, 1: right
            self.reset_balloon()

    def config_adjust(self, name, base, default_value):
        path, config = self.balloon[self.balloon_id]
        value = config.get_with_type(name, int)
        if value is None:
            value = self.desc.get_with_type(name, int)
        if value is None:
            value = default_value
        if value < 0:
            value = base + value
        return int(value * self.scale / 100)

    def __get_with_scaling(self, name, conv, default_value):
        path, config = self.balloon[self.balloon_id]
        value = config.get_with_type(name, conv)
        if value is None:
            value = self.desc.get_with_type(name, conv)
            if value is None:
                value = default_value
        return conv(value * self.scale / 100)

    def __move(self):
        x, y = self.get_position()
        self.window.resize_move(x, y)

    def set_position(self, base_x, base_y):
        if self.balloon_id is None: ## FIXME
            return
        px, py = self.get_balloon_windowposition()
        if self.direction == 0:
            w, h = self.get_balloon_size()
            x = base_x + px - w
        else:
            x = base_x + px
        y = base_y + py
        self.position = (x, y)
        if self.__shown:
            self.__move()

    def get_position(self):
        return self.position

    def destroy(self, finalize=0):
        self.window.remove(self.darea)
        self.darea.destroy()
        self.window.destroy()

    def is_shown(self):
        return 1 if self.__shown else 0

    def show(self):
        if self.__shown:
            return
        self.__shown = 1
        self.darea.show()
        self.window.show()
        # make sure window is in its position
        self.__move()
        self.raise_()

    def hide(self):
        if not self.__shown:
            return
        self.window.hide()
        self.__shown = 0
        self.images = []

    def raise_(self):
        if self.__shown:
            self.window.window.raise_()

    def lower(self):
        if self.__shown:
            self.window.window.lower()

    def show_sstp_message(self, message, sender):
        if self.sstp_region is None:
            self.show()
        self.sstp_message = u'{0} ({1})'.format(message, sender)
        x, y, w, h = self.sstp_region
        self.sstp_layout.set_text(self.sstp_message)
        message_width, message_height = self.sstp_layout.get_pixel_size()
        if message_width > w:
            self.sstp_message = u'... ({0})'.format(sender)
            i = 0
            while 1:
                i += 1
                s = u'{0}... ({1})'.format(message[:i], sender)
                self.sstp_layout.set_text(s)
                message_width, message_height = \
                               self.sstp_layout.get_pixel_size()
                if message_width > w:
                    break
                self.sstp_message = s
        self.darea.queue_draw()

    def hide_sstp_message(self):
        self.sstp_message = None
        self.darea.queue_draw()

    def redraw_sstp_message(self):
        if self.sstp_message is None:
            return
        cr = self.darea.window.cairo_create()
        # draw sstp marker
        if self.sstp_pixbuf is not None:
            x, y = self.sstp[0]
            cr.set_source_pixbuf(self.sstp_pixbuf[0], x, y)
            cr.paint()
        # draw sstp message
        x, y, w, h = self.sstp_region
        self.sstp_layout.set_text(self.sstp_message)
        cr.set_source_rgba(*self.text_normal_color)
        cr.move_to(x, y)
        cr.show_layout(self.sstp_layout)
        del cr

    def redraw_arrow0(self):
        if self.lineno <= 0:
            return
        cr = self.darea.window.cairo_create()
        x, y = self.arrow[0]
        cr.set_source_pixbuf(self.arrow0_pixbuf[0], x, y)
        cr.paint()
        del cr

    def redraw_arrow1(self):
        if self.lineno + self.lines >= len(self.text_buffer):
            return
        cr = self.darea.window.cairo_create()
        x, y = self.arrow[1]
        cr.set_source_pixbuf(self.arrow1_pixbuf[0], x, y)
        cr.paint()
        del cr

    def set_markup(self, index, text):
        tags_ = ('sup', 'sub', 's', 'u')
        count_ = {}
        for tag_ in tags_:
            count_[tag_] = 0
        markup_list = []
        for sl, sn, tag in self.meta_buffer:
            if sl == index:
                markup_list.append((sn, tag))
        if not markup_list:
            return glib.markup_escape_text(text)
        markup_list.sort()
        markup_list.reverse()
        pn = len(text)
        for sn, tag in markup_list:
            text = ''.join((text[:sn], tag,
                            glib.markup_escape_text(text[sn:pn]), text[pn:]))
            pn = sn
            if tag[1] == '/':
                tag_ = tag[2:-1]
                assert tag_ in tags_
                count_[tag_] -= 1
                if count_[tag_] < 0:
                    text = ''.join(('<', tag_, '>', text))
                    count_[tag_] += 1
            else:
                tag_ = tag[1:-1]
                assert tag_ in tags_
                count_[tag_] += 1
                if count_[tag_] > 0:
                    text = ''.join((text, '</', tag_, '>'))
                    count_[tag_] -= 1
        return text

    def redraw(self, darea, event):
        if not self.__shown:
            return True
        assert self.balloon_pixbuf is not None
        cr = darea.window.cairo_create()
        cr.save()
        cr.set_operator(cairo.OPERATOR_CLEAR)
        cr.paint()
        cr.restore()
        cr.set_source_pixbuf(self.balloon_pixbuf[0], 0, 0)
        cr.paint_with_alpha(self.alpha_channel)
        # draw images
        for i in range(len(self.images)):
            pixbuf, (w, h), (x, y) = self.images[i]
            scale = self.scale
            w = max(8, int(w * scale / 100))
            h = max(8, int(h * scale / 100))
            if x == 'centerx':
                bw, bh = self.get_balloon_size()
                x = (bw - w) / 2
            else:
                try:
                    x = int(x)
                except:
                    continue
            if y == 'centery':
                bw, bh = self.get_balloon_size()
                y = (bh - h) / 2
            else:
                try:
                    y = int(y)
                except:
                    continue
            pixbuf = pixbuf.scale_simple(w, h, gtk.gdk.INTERP_BILINEAR)
            cr.set_source_pixbuf(pixbuf, x, y)
            cr.paint_with_alpha(self.alpha_channel)
        # draw text
        i = self.lineno
        j = len(self.text_buffer)
        line = 0
        while line < self.lines:
            if i >= j:
                break
            x, y, w, h = self.line_regions[line]
            if self.text_buffer[i].endswith('\n[half]'):
                new_y = int(y + (self.font_height + self.line_space) / 2)
                markup = self.set_markup(i, self.text_buffer[i][:-7])
            else:
                new_y = int(y + self.font_height + self.line_space)
                markup = self.set_markup(i, self.text_buffer[i])
            self.update_line_regions(line + 1, new_y)
            self.layout.set_markup(markup)
            cr.set_source_rgba(*self.text_normal_color)
            cr.move_to(x, y)
            cr.show_layout(self.layout)
            if self.sstp_pixbuf is not None:
                for l, c in self.sstp_marker:
                    if l == i:
                        mw, mh = self.sstp_pixbuf[1]
                        self.layout.set_text(self.text_buffer[i][:c])
                        text_w, text_h = self.layout.get_pixel_size()
                        mx = x + text_w
                        my = y + (self.font_height + self.line_space) / 2
                        my = my - mh / 2
                        cr.set_source_pixbuf(self.sstp_pixbuf[0], mx, my)
                        cr.paint_with_alpha(self.alpha_channel)
            i += 1
            line += 1
        del cr
        if self.side == 0 and self.sstp_message:
            self.redraw_sstp_message()
        if self.selection is not None:
            self.update_link_region(darea, self.selection)
        self.redraw_arrow0()
        self.redraw_arrow1()
        return False

    def update_link_region(self, darea, index):
        cr = darea.window.cairo_create()
        sl = self.link_buffer[index][0]
        el = self.link_buffer[index][2]
        if self.lineno <= sl <= self.lineno + self.lines:
            sn = self.link_buffer[index][1]
            en = self.link_buffer[index][3]
            for n in range(sl, el + 1):
                if n - self.lineno >= len(self.line_regions):
                    break
                x, y, w, h = self.line_regions[n - self.lineno]
                if sl == el:
                    markup = self.set_markup(n, self.text_buffer[n][:sn])
                    self.layout.set_markup(markup)
                    text_w, text_h =  self.layout.get_pixel_size()
                    x += text_w
                    markup = self.set_markup(n, self.text_buffer[n][sn:en])
                    self.layout.set_markup(markup)
                    text_w, text_h =  self.layout.get_pixel_size()
                    w = text_w
                    start = sn
                    end = en
                elif n == sl:
                    markup = self.set_markup(n, self.text_buffer[n][:sn])
                    self.layout.set_markup(markup)
                    text_w, text_h =  self.layout.get_pixel_size()
                    x += text_w
                    markup = self.set_markup(n, self.text_buffer[n][sn:])
                    self.layout.set_markup(markup)
                    text_w, text_h =  self.layout.get_pixel_size()
                    w = text_w
                    start = sn
                    end = len(self.text_buffer[n])
                elif n == el:
                    markup = self.set_markup(n, self.text_buffer[n][:en])
                    self.layout.set_markup(markup)
                    text_w, text_h =  self.layout.get_pixel_size()
                    w = text_w
                    start = 0
                    end = en
                else:
                    markup = self.set_markup(n, self.text_buffer[n])
                    self.layout.set_markup(markup)
                    text_w, text_h =  self.layout.get_pixel_size()
                    w = text_w
                    start = 0
                    end = len(self.text_buffer[n])
                markup = self.set_markup(n, self.text_buffer[n][start:end])
                self.layout.set_markup(markup)
                cr.set_source_rgba(*self.cursor_color)
                cr.rectangle(x, y, w, h)
                cr.fill()
                cr.move_to(x, y)
                cr.set_source_rgba(*self.text_active_color)
                cr.show_layout(self.layout)
        del cr

    def check_link_region(self, px, py):
        new_selection = None
        for i in range(len(self.link_buffer)):
            sl = self.link_buffer[i][0]
            el = self.link_buffer[i][2]
            if self.lineno <= sl <= self.lineno + self.lines:
                sn = self.link_buffer[i][1]
                en = self.link_buffer[i][3]
                for n in range(sl,el + 1):
                    if n - self.lineno >= len(self.line_regions):
                        break
                    x, y, w, h = self.line_regions[n - self.lineno]
                    if n == sl:
                        markup = self.set_markup(n, self.text_buffer[n][:sn])
                        self.layout.set_markup(markup)
                        text_w, text_h =  self.layout.get_pixel_size()
                        x += text_w
                    if n == sl and n == el:
                        markup = self.set_markup(n, self.text_buffer[n][sn:en])
                    elif n == el:
                        markup = self.set_markup(n, self.text_buffer[n][:en])
                    else:
                        markup = self.set_markup(n, self.text_buffer[n])
                    self.layout.set_markup(markup)
                    text_w, text_h =  self.layout.get_pixel_size()
                    w = text_w
                    if x <= px < x + w and y <= py < y + h:
                        new_selection = i
                        break
        if new_selection is not None:
            if self.selection != new_selection:
                sl, sn, el, en, link_id, raw_text, text = \
                    self.link_buffer[new_selection]
                self.request_parent(
                    'NOTIFY', 'notify_event',
                    'OnChoiceEnter', raw_text, link_id, self.selection)
        else:
            if self.selection is not None:
                self.request_parent('NOTIFY', 'notify_event', 'OnChoiceEnter')
        if new_selection == self.selection:
            return 0
        else:
            self.selection = new_selection
            return 1 # dirty flag

    def motion_notify(self, darea, event):
        if event.is_hint:
            x, y, state = self.darea.window.get_pointer()
        else:
            x, y, state = event.x, event.y, event.state
        if not self.link_buffer:
            return True
        px = int(x)
        py = int(y)
        if self.check_link_region(px, py):
            self.darea.queue_draw()
        return True

    def scroll(self, darea, event):
        px = int(event.x)
        py = int(event.y)
        if event.direction == gtk.gdk.SCROLL_UP:
            if self.lineno > 0:
                self.lineno = max(self.lineno - 2, 0)
                self.check_link_region(px, py)
                self.darea.queue_draw()
        elif event.direction == gtk.gdk.SCROLL_DOWN:
            if self.lineno + self.lines < len(self.text_buffer):
                self.lineno = min(self.lineno + 2,
                                  len(self.text_buffer) - self.lines)
                self.check_link_region(px, py)
                self.darea.queue_draw()
        return True

    def button_press(self, darea, event):
        self.request_parent('NOTIFY', 'reset_idle_time')
        click = 1 if event.type == gtk.gdk.BUTTON_PRESS else 2
        if self.request_parent('GET', 'is_paused'):
            self.request_parent('NOTIFY', 'notify_balloon_click',
                                event.button, click, self.side)
            return True
        # arrows
        px = int(event.x)
        py = int(event.y)
        # up arrow
        w, h = self.arrow0_pixbuf[1]
        x, y = self.arrow[0]
        if x <= px <= x + w and y <= py <= y + h:
            if self.lineno > 0:
                self.lineno = max(self.lineno - 2, 0)
                self.darea.queue_draw()
            return True
        # down arrow
        w, h = self.arrow1_pixbuf[1]
        x, y = self.arrow[1]
        if x <= px <= x + w and y <= py <= y + h:
            if self.lineno + self.lines < len(self.text_buffer):
                self.lineno = min(self.lineno + 2,
                                  len(self.text_buffer) - self.lines)
                self.darea.queue_draw()
            return True
        # links
        if self.selection is not None:
            sl, sn, el, en, link_id, raw_text, text = \
                self.link_buffer[self.selection]
            self.request_parent('NOTIFY', 'notify_link_selection',
                                link_id, raw_text, self.selection)
            return True
        # balloon's background
        self.request_parent('NOTIFY', 'notify_balloon_click',
                            event.button, click, self.side)
        return True

    def clear_text(self):
        self.selection = None
        self.lineno = 0
        self.text_buffer = []
        self.meta_buffer = []
        self.link_buffer = []
        self.newline_required = 0
        self.images = []
        self.sstp_marker = []
        self.darea.queue_draw()

    def get_text_count(self):
        return self.text_count

    def reset_text_count(self):
        self.text_count = 0

    def set_newline(self):
        self.newline_required = 1

    def append_text(self, text):
        if not self.text_buffer:
            s = ''
            column = 0
            index = 0
        elif self.newline_required:
            s = ''
            column = 0
            self.newline_required = 0
            index = len(self.text_buffer)
        else:
            index = len(self.text_buffer) - 1
            s = self.text_buffer.pop(-1)
            column = len(s)
        i = len(s)
        text = ''.join((s, text))
        j = len(text)
        self.text_count += j
        p = 0
        while 1:
            if i >= j:
                self.text_buffer.append(text[p:i])
                self.draw_last_line(column)
                break
            if text[i] == '\n':
                if j >= i + 7 and text[i:i + 7] == '\n[half]':
                    self.text_buffer.append(''.join((text[p:i], '\n[half]')))
                    p = i = i + 7
                else:
                    self.text_buffer.append(text[p:i])
                    p = i = i + 1
                self.draw_last_line(column)
                column = 0
                continue
            n = i + 1
            if not self.__shown:
                self.show()
            markup = self.set_markup(index, text[p:n])
            self.layout.set_markup(markup)
            text_width, text_height =  self.layout.get_pixel_size()
            if text_width > self.line_width:
                self.text_buffer.append(text[p:i])
                self.draw_last_line(column)
                column = 0
                p = i
            i = n

    def append_sstp_marker(self):
        if self.sstp_pixbuf is None:
            return
        if not self.text_buffer:
            line = 0
            offset = 0
        else:
            line = len(self.text_buffer) - 1
            offset = len(self.text_buffer[-1])
        if self.newline_required:
            line = line + 1
            offset = 0
        self.sstp_marker.append((line, offset))
        w, h = self.sstp_pixbuf[1]
        i = 1
        while 1:
            space = u'\u3000' * i ## FIXME
            self.layout.set_text(space)
            text_w, text_h = self.layout.get_pixel_size()
            if text_w > w:
                break
            else:
                i += 1
        self.append_text(space)
        self.draw_last_line(offset)

    def append_link_in(self, link_id):
        if not self.text_buffer:
            sl = 0
            sn = 0
        else:
            sl = len(self.text_buffer) - 1
            sn = len(self.text_buffer[-1])
        self.link_buffer.append((sl, sn, sl, sn, link_id, '', ''))

    def append_link_out(self, link_id, text):
        if not text:
            return
        raw_text = text
        if not self.text_buffer:
            el = 0
            en = 0
        else:
            el = len(self.text_buffer) - 1
            en = len(self.text_buffer[-1])
        for i in range(len(self.link_buffer)):
            if self.link_buffer[i][4] == link_id:
                sl = self.link_buffer[i][0]
                sn = self.link_buffer[i][1]
                self.link_buffer.pop(i)
                self.link_buffer.insert(
                    i, (sl, sn, el, en, link_id, raw_text, text))
                break

    def append_meta(self, tag):
        if not tag:
            return
        if not self.text_buffer:
            sl = 0
            sn = 0
        else:
            sl = len(self.text_buffer) - 1
            sn = len(self.text_buffer[-1])
        self.meta_buffer.append((sl, sn, tag))

    def append_image(self, path, x, y):
        try:
            pixbuf = ninix.pix.create_pixbuf_from_file(path)
        except:
            return
        self.show()
        w = pixbuf.get_width()
        h = pixbuf.get_height()
        self.images.append((pixbuf, (w, h), (x, y)))
        self.darea.queue_draw()

    def draw_last_line(self, column=0):
        if not self.__shown:
            return
        line = len(self.text_buffer) - 1
        if self.lineno <= line < self.lineno + self.lines:
            x, y, w, h = self.line_regions[line - self.lineno]
            if self.text_buffer[line].endswith('\n[half]'):
                offset = line - self.lineno + 1
                new_y = int(y + (self.font_height + self.line_space) / 2)
                self.update_line_regions(offset, new_y)
            else:
                self.darea.queue_draw()
            if self.sstp_pixbuf is not None:
                for l, c in self.sstp_marker:
                    if l == line:
                        mw, mh = self.sstp_pixbuf[1]
                        self.layout.set_text(self.text_buffer[l][:c])
                        text_w, text_h = self.layout.get_pixel_size()
                        mx = x + text_w
                        my = y + (self.font_height + self.line_space) / 2
                        my = my - mh / 2
                        cr = self.darea.window.cairo_create()
                        cr.set_source_pixbuf(self.sstp_pixbuf[0], mx, my)
                        cr.paint()
                        del cr
        else:
            self.darea.queue_draw()
            while line >= self.lineno + self.lines:
                self.lineno += 1
                self.darea.queue_draw()


class CommunicateWindow(object):

    NAME = ''
    ENTRY = ''

    def __init__(self):
        self.request_parent = lambda *a: None # dummy
        self.window = None

    def set_responsible(self, request_method):
        self.request_parent = request_method

    def new(self, desc, balloon):
        if self.window:
            self.window.destroy()
        self.window = gtk.Window()
        self.window.set_title('communicate')
        self.window.set_decorated(False)
        self.window.set_resizable(False)
        self.window.connect('delete_event', self.delete)
        self.window.connect('key_press_event', self.key_press)
        self.window.connect('button_press_event', self.button_press)
        self.window.set_events(gtk.gdk.BUTTON_PRESS_MASK)
        self.window.set_modal(True)
        self.window.set_position(gtk.WIN_POS_CENTER)
        self.window.realize()
        w = desc.get_with_type('communicatebox.width', int, 250)
        h = desc.get_with_type('communicatebox.height', int, -1)
        self.entry = gtk.Entry()
        self.entry.connect('activate', self.activate)
        self.entry.set_size_request(w, h)
        self.entry.show()
        pixbuf = None
        if balloon:
            path, config = balloon
            # load pixbuf
            try:
                pixbuf = ninix.pix.create_pixbuf_from_file(path)
            except:
                pixbuf = None
        if pixbuf is not None:
            darea = gtk.DrawingArea()
            darea.set_events(gtk.gdk.EXPOSURE_MASK)
            darea.connect('expose_event', self.redraw, pixbuf)
            darea.show()
            x = desc.get_with_type('communicatebox.x', int, 10)
            y = desc.get_with_type('communicatebox.y', int, 20)
            fixed = gtk.Fixed()
            fixed.put(darea, 0, 0)
            fixed.put(self.entry, x, y)
            fixed.show()
            self.window.add(fixed)
            w = pixbuf.get_width()
            h = pixbuf.get_height()
            darea.set_size_request(w, h)
            mask = gtk.gdk.Pixmap(None, w, h, 1)
            pixbuf.render_threshold_alpha(mask, 0, 0, 0, 0, w, h, 1)
            self.window.shape_combine_mask(mask, 0, 0)
        else:
            box = gtk.HBox(spacing=10)
            box.set_border_width(10)
            if self.ENTRY:
                label = gtk.Label(self.ENTRY)
                box.pack_start(label, False)
                label.show()
            box.pack_start(self.entry)
            self.window.add(box)
            box.show()

    def redraw(self, darea, event, pixbuf):
        cr = darea.window.cairo_create()
        cr.save()
        cr.set_operator(cairo.OPERATOR_CLEAR)
        cr.paint()
        cr.restore()
        cr.set_source_pixbuf(pixbuf, 0, 0)
        cr.paint()
        del cr

    def destroy(self):
        if self.window:
            self.window.destroy()
            self.window = None

    def delete(self, widget, event):
        self.window.hide()
        self.cancel()
        return True

    def key_press(self, widget, event):
        if event.keyval == gtk.keysyms.Escape:
            self.window.hide()
            self.cancel()
            return True
        return False

    def button_press(self, widget, event):
        if event.button in [1, 2]:
            self.window.begin_move_drag(
                event.button, int(event.x_root), int(event.y_root),
                gtk.get_current_event_time())
        return True

    def activate(self, widget):
        self.window.hide()
        self.enter()
        return True

    def show(self, default=''):
        self.entry.set_text(default)
        self.window.show()

    def enter(self):
        pass

    def cancel(self):
        pass


class CommunicateBox(CommunicateWindow):

    NAME = 'communicatebox'
    ENTRY = 'Communicate'

    def new(self, desc, balloon):
        CommunicateWindow.new(self, desc, balloon)
        self.window.set_modal(False)

    def delete(self, widget, event):
        self.window.hide()
        self.cancel()
        self.request_parent('NOTIFY', 'reset_user_interaction')
        return True

    def key_press(self, widget, event):
        if event.keyval == gtk.keysyms.Escape:
            self.window.hide()
            self.cancel()
            self.request_parent('NOTIFY', 'reset_user_interaction')
            return True
        return False

    def activate(self, widget):
        self.enter()
        self.entry.set_text('')
        return True

    def enter(self):
        self.send(self.entry.get_text())

    def cancel(self):
        self.send(None)

    def send(self, data):
        if data:
            data = unicode(data, 'utf-8')
        if data is not None:
            self.request_parent('NOTIFY', 'notify_event',
                                'OnCommunicate', 'user', data)


class TeachBox(CommunicateWindow):

    NAME = 'teachbox'
    ENTRY = 'Teach'

    def enter(self):
        self.send(self.entry.get_text())

    def cancel(self):
        self.send(None)

    def send(self, data):
        if data:
            data = unicode(data, 'utf-8')
        self.request_parent('NOTIFY', 'notify_user_teach', data)
        self.request_parent('NOTIFY', 'reset_user_interaction')


class InputBox(CommunicateWindow):

    NAME = 'inputbox'
    ENTRY = 'Input'

    def new(self, desc, balloon):
        CommunicateWindow.new(self, desc, balloon)
        self.symbol = None
        self.limittime = -1

    def set_symbol(self, symbol):
        self.symbol = symbol
        
    def set_limittime(self, limittime):
        try:
            limittime = int(limittime)
        except ValueError:
            limittime = -1
        self.limittime = limittime

    def show(self, default):
        if default is not None:
            try:
                text = unicode(default).encode('utf-8')
            except:
                text = ''
        else:
            text = ''
        if int(self.limittime) < 0:
            self.timeout_id = None
        else:
            self.timeout_id = glib.timeout_add(self.limittime, self.timeout)
        CommunicateWindow.show(self, text)

    def timeout(self):
        self.window.hide()
        self.send('timeout')

    def enter(self):
        self.send(self.entry.get_text())

    def cancel(self):
        self.send(None)

    def send(self, data):
        if self.timeout_id is not None:
            glib.source_remove(self.timeout_id)            
        if data:
            data = unicode(data, 'utf-8')
        if data is None:
            data = ''
        ## CHECK: symbol
        if self.symbol == 'OnUserInput' and \
           self.request_parent('GET', 'notify_event', 'OnUserInput', data):
            pass
        elif self.request_parent('GET', 'notify_event', 'OnUserInput', self.symbol, data):
            pass
        elif self.request_parent('GET', 'notify_event', self.symbol, data):
            pass
        self.request_parent('NOTIFY', 'reset_user_interaction')


def test():
    pass

if __name__ == '__main__':
    test()
